#include "iptrap.h"
#include "iptrap_p.h"
#ifdef WITH_DMALLOC
# include <dmalloc.h>
#endif

static inline void reuseaddr(const int fd)
{
	static const int on = 1;
	(void) setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on);
}

/*
 * Check it an address is valid, return 1 if ok, 0 otherwise.
 */

static int checkvalidaddr(const struct sockaddr_storage * const addr)
{
	if (addr == NULL) {
		return 0;
	}
	if (STORAGE_FAMILY(*addr) == AF_INET6) {
		if (IN6_IS_ADDR_MULTICAST(&STORAGE_SIN_ADDR6_NF(*addr)) ||
			IN6_IS_ADDR_UNSPECIFIED(&STORAGE_SIN_ADDR6_NF(*addr))) {
			return 0;
		}
		return 1;
	} else if (STORAGE_FAMILY(*addr) == AF_INET) {
		if (ntohl(STORAGE_SIN_ADDR(*addr)) == INADDR_ANY ||
			ntohl(STORAGE_SIN_ADDR(*addr)) == INADDR_NONE ||
			ntohl(STORAGE_SIN_ADDR(*addr)) == INADDR_BROADCAST ||
			IN_MULTICAST(ntohl(STORAGE_SIN_ADDR(*addr))) ||
			IN_BADCLASS(ntohl(STORAGE_SIN_ADDR(*addr)))) {
			return 0;
		}
		return 1;
	}
	return 0;
}

/* 
 * Convert a 4-in-6 address in pure IPv4.
 */

static void fourinsix(struct sockaddr_storage *v6)
{
	struct sockaddr_storage v4;
		
	if (STORAGE_FAMILY(*v6) != AF_INET6 ||
		IN6_IS_ADDR_V4MAPPED(&STORAGE_SIN_ADDR6_NF(*v6)) == 0) {
		return;
	}
	memset(&v4, 0, sizeof v4);
	STORAGE_FAMILY(v4) = AF_INET;
	memcpy(&STORAGE_SIN_ADDR(v4),
		   (unsigned char *) &STORAGE_SIN_ADDR6(*v6) + 12,
		   sizeof STORAGE_SIN_ADDR(v4));
	STORAGE_PORT(v4) = STORAGE_PORT6(*v6);
#ifdef HAVE_SS_LEN
	STORAGE_LEN(v4) = sizeof(struct sockaddr_in);
#endif
	*v6 = v4;
}		

/*
 * check for ipv6 support in kernel.
 */

static void check_ipv6_support(void)
{
	int p;
	
	if ((p = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP)) != -1) {
		close(p);
		v6ready++;
	}
}

/*
 * Set the close-on-exec flag on a socket descriptor.
 */

static inline void setcloexec(const int fd)
{
	fcntl(fd, F_SETFD, FD_CLOEXEC);
}

/*
 * Register a new ip/port : create a socket, bind it, listen to it,
 * and add it to the poll.
 */

static int addport(const char * const ip, const char * const port)
{
	struct addrinfo hints, *res;
	int rc;
	struct sockaddr_storage addr;
	socklen_t addrlen;
	int family;
	int fd;
	struct pollfd *polls_cur;

	if (nb_sockets >= (sizeof polls / (sizeof polls[0]))) {
		return -1;
	}
	(void) memset(&hints, 0, sizeof hints);	
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = IPPROTO_TCP;
	hints.ai_addr = NULL;	
	hints.ai_family = AF_UNSPEC;
	if ((rc = getaddrinfo(ip, port, &hints, &res)) != 0) {
		fprintf(stderr, "getaddrinfo : %s\n", gai_strerror(rc));
		return -1;
	}	
	if (ip == NULL) {
		if (v6ready) {
			STORAGE_SIN_ADDR6_NF(*res->ai_addr) = in6addr_any;
			res->ai_family = PF_INET6;
			res->ai_addrlen = sizeof(struct sockaddr_in6); 
		} else {
			STORAGE_SIN_ADDR(*res->ai_addr) = htonl(INADDR_ANY);
			res->ai_family = PF_INET;		
			res->ai_addrlen = sizeof(struct sockaddr_in);
		}
	}
	family = res->ai_family;
	addrlen = res->ai_addrlen;
	(void) memcpy(&addr, res->ai_addr, addrlen);
	freeaddrinfo(res);
	if ((fd = socket(family, SOCK_STREAM, IPPROTO_TCP)) < 0) {
		perror("socket");
		return -1;
	}
	reuseaddr(fd);
	setcloexec(fd);
	if (bind(fd, (struct sockaddr *) &addr, addrlen) < 0) {
		perror("bind");
		(void) close(fd);
		return -1;
	}
	if (listen(fd, DEFAULT_BACKLOG) < 0) {
		perror("listen");
		(void) close(fd);
		return -1;
	}
	polls_cur = &polls[nb_sockets];
	nb_sockets++;	
	polls_cur->fd = fd;
	polls_cur->events = POLLIN;
	polls_cur->revents = 0;
	
	return 0;
}

/*
 * Call iptables/iptables6/ipchains to firewall an IP address.
 */

static void filterclient(const struct sockaddr_storage * const client_addr,
						 const socklen_t client_addr_len,
						 const unsigned short localport)
{
	pid_t pid;
	char hbuf[NI_MAXHOST];		
	
	if (getnameinfo((struct sockaddr *) client_addr,
					client_addr_len, hbuf, sizeof hbuf,
					NULL, 0, NI_NUMERICHOST) == 0) {
		syslog(LOG_ALERT,
			   "[%s] connected to unauthorized port [%u] - firewalling", 
			   hbuf, (unsigned int) localport);
		pid = fork();
		if (pid == (pid_t) 0) {
			if (external_script != NULL) {
				char *envar;
				size_t envar_size;
				
				envar_size = sizeof TCPREMOTEIP + 1U + sizeof hbuf;
				if ((envar = malloc(envar_size)) == NULL ||
					snprintf(envar, envar_size, 
							 TCPREMOTEIP "=%s", hbuf) < 0) {
					_exit(EXIT_FAILURE);
				}
				(void) putenv(envar);
				if (localport > 0U) {
					envar_size = sizeof TCPLOCALPORT + 1U + sizeof "65535";
					if ((envar = malloc(envar_size)) != NULL &&
						snprintf(envar, envar_size, TCPLOCALPORT "=%u", 
								 (unsigned int) localport) > 0) {
						(void) putenv(envar);
					}
				}
				execl(external_script, external_script, NULL);
				free(envar);
			} else {
#ifdef IPTABLES
				if (STORAGE_FAMILY(*client_addr) == AF_INET6) {
# ifdef IP6TABLES
					execl(IP6TABLES, IP6TABLES, "-I", chain,
						  "--source", hbuf, "--jump", jump, NULL);
# endif
				} else {
					execl(IPTABLES, IPTABLES, "-I", chain,
						  "--source", hbuf, "--jump", jump, NULL);
				}
#elif defined(IPCHAINS)
				if (STORAGE_FAMILY(*client_addr) != AF_INET6) {
					execl(IPCHAINS, IPCHAINS, "-I", chain,
						  "--source", hbuf, "--jump", jump, NULL);
				}
#endif
			}
			_exit(EXIT_FAILURE);
		} else if (pid != (pid_t) -1) {
#ifdef HAVE_WAITPID
			(void) waitpid(pid, NULL, 0);
#else
			while (wait3(NULL, 0, NULL) != pid);
#endif
		}
	}	
}

/*
 * Check if an IP address shouldn't be banned.
 */

static int ismasked(const struct sockaddr_storage * const client_addr)
{
	register const IPMask *ipmasks_scan = ipmasks;
	unsigned nb_masks_cnt = nb_masks;

	if (STORAGE_FAMILY(*client_addr) == AF_INET6) {
		if (memcmp(&in6addr_loopback, STORAGE_SIN_ADDR6(*client_addr),
				   sizeof in6addr_loopback) == 0) {
			return 1;
		}
		while (nb_masks_cnt > 0U) {			
			if (ipmasks_scan->family == AF_INET6 &&
				memcmp(&ipmasks_scan->ipmask6.addr,
					   STORAGE_SIN_ADDR6(*client_addr),
					   sizeof ipmasks_scan->ipmask6.addr) == 0) {
				return 1;
			}
			ipmasks_scan++;
			nb_masks_cnt--;
		}
	} else {
		if (htonl(INADDR_LOOPBACK) == STORAGE_SIN_ADDR(*client_addr)) {
			return 1;
		}
		while (nb_masks_cnt > 0U) {
			if (ipmasks_scan->family != AF_INET6 &&
				ipmasks_scan->ipmask4.addr ==
				(STORAGE_SIN_ADDR(*client_addr) &
				 ipmasks_scan->ipmask4.mask)) {
				return 1;
			}
			ipmasks_scan++;
			nb_masks_cnt--;
		}		
	}
	
	return 0;
}

/*
 * Find the local port a client connected to.
 */

static unsigned short getlocalport(const int fd)
{
	struct sockaddr_storage local_sockaddr;
	socklen_t local_sockaddr_len;
	
	local_sockaddr_len = sizeof local_sockaddr;
	if (getsockname(fd, (struct sockaddr *) &local_sockaddr,
					&local_sockaddr_len) < 0) {
		return ntohs(0U);
	}
	if (STORAGE_FAMILY(local_sockaddr) == AF_INET) {
		return ntohs(STORAGE_PORT(local_sockaddr));
	}
	if (STORAGE_FAMILY(local_sockaddr) == AF_INET6) {
		return ntohs(STORAGE_PORT6(local_sockaddr));
	}
	return ntohs(0U);
}

/*
 * Wait for a new client, hang up the connection and call filterclient() .
 */

static void waitclient(void)
{
	struct sockaddr_storage client_addr;
	socklen_t client_addr_len;	
	int client;
	register struct pollfd *polls_scan;
	unsigned int nb_sockets_cnt;
	unsigned short localport;
	
	while (poll(polls, nb_sockets, -1) < 0 && errno == EINTR);
	polls_scan = polls;
	nb_sockets_cnt = nb_sockets;
	do {
		if ((polls_scan->revents & POLLIN) == 0) {
			goto next;
		}
		polls_scan->revents = 0;
		(void) memset(&client_addr, 0, sizeof client_addr);
		client_addr_len = sizeof client_addr;		
		client = accept(polls_scan->fd,
						(struct sockaddr *) &client_addr,
						(socklen_t *) &client_addr_len);
		if (client < 0) {
			goto next;
		}
		localport = getlocalport(client);		
		(void) shutdown(client, 2);
		(void) close(client);
		if (checkvalidaddr(&client_addr) == 0) {
			goto next;
		}
		fourinsix(&client_addr);
		if (ismasked(&client_addr) != 0) {
			goto next;
		}		
		filterclient(&client_addr, client_addr_len, localport);
		next:
		polls_scan++;
		nb_sockets_cnt--;
	} while (nb_sockets_cnt > 0U);	
}

static void usage(void)
{
	puts("\nUsage :\n\n"
		 "iptrap -p [ip:]port [-p [ip:]port] [...] [-x ip] [-x ip] [...]\n"
		 "       [-c chain] [-j target rule] [-s /path/to/script] [-B]\n\n"
#ifdef IPTABLES		 
		 "Launched command : " IPTABLES "\n"
#elif defined (IPCHAINS)
		 "Launched command : " IPCHAINS "\n"
#endif
		 );
}

/*
 * Parse command-line options.
 */

static int parseoptions(const int argc, char *argv[])
{
	int fodder;
	
	while ((fodder = getopt(argc, argv, "hp:x:c:j:s:B")) != -1) {
		switch(fodder) {
		case 'p' :
			{
				char *dlm;
				const char *port;
				const char *ip;
				
				if ((dlm = strrchr(optarg, ':')) == NULL) {	
					ip = NULL;
					port = optarg;
				} else {
					*dlm = 0;
					ip = optarg;
					port = dlm + 1;
				}
				(void) addport(ip, port);
			}
			break;
		case 'x' :
			{
				struct addrinfo hints, *res;
				int rc;
				IPMask new_mask;
							
				if (nb_masks >= (sizeof ipmasks / (sizeof ipmasks[0]))) {
					break;
				}
				(void) memset(&hints, 0, sizeof hints);
				hints.ai_addr = NULL;
				if ((rc = getaddrinfo(optarg, NULL, &hints, &res)) == 0) {
					if (res->ai_family == AF_INET6) {
						new_mask.ipmask6.family = AF_INET6;
						(void) memcmp(new_mask.ipmask6.addr,
									  STORAGE_SIN_ADDR6(*res->ai_addr),
									  sizeof new_mask.ipmask6.addr);
					} else {
						new_mask.ipmask4.family = AF_INET;
						new_mask.ipmask4.addr = STORAGE_SIN_ADDR(*res->ai_addr);
						new_mask.ipmask4.mask = ~0UL;
					}
					ipmasks[nb_masks] = new_mask;
					nb_masks++;
				}
			}
			break;
		case 'c' :
			if ((chain = strdup(optarg)) == NULL) {
				return -1;
			}
			break;
		case 'j' :
			if ((jump = strdup(optarg)) == NULL) {
				return -1;
			}
			break;
		case 's' :
			if ((external_script = strdup(optarg)) == NULL) {
				return -1;
			}
			break;
		case 'B' :
			daemonize = 1;
			break;
		default :
			usage();
			return 1;
		}
	}
	return 0;
}

/*
 * Detach the process.
 */
	
static void dodaemonize(void)
{
	pid_t child;
	
	if (daemonize) {
		if ((child = fork()) < (pid_t) 0) {
			perror("Daemonization failed - fork");
			syslog(LOG_ERR, "Daemonization failed : [fork: %s]",
				   strerror(errno));
			return;
		} else if (child != (pid_t) 0) {
			_exit(EXIT_SUCCESS);
		} else {
			if (setsid() == (pid_t) -1) {
				syslog(LOG_ERR, "Daemonization failed : [setsid: %s]",
					   strerror(errno));
			}
			chdir("/");
			if (isatty(1)) {
				close(1);
			}
			if (isatty(2)) {
				close(2);
			}
		}
	}	
}

int main(int argc, char *argv[])
{	
	close(0);
	check_ipv6_support();
	openlog("iptrap", 0, LOG_AUTH);
	if (parseoptions(argc, argv) != 0) {
		return -1;
	}
	if (nb_sockets == 0U) {
		usage();
		return 1;
	}
	dodaemonize();
	syslog(LOG_INFO, "IPtrap starting. Chain : [%s] Target : [%s]",
		   chain, jump);
	for (;;) {
		waitclient();
	}
	closelog();
		
	return 0;	
}

